#!/usr/bin/ruby

require 'multi'
require 'smulti'

module SExpressionParser
  class Base
    def initialize(text, *start)
      @list = start
      parse(text)
    end
    
    def add(item)
      @list.push(item)
    end

    def leave(text)
      @unwanted = text
    end

    def unwanted
      return @unwanted
    end
    
    def add_parse(item, rest)
      add(item)
      return parse(rest)
    end

    def add_leave(item, rest)
      add(item)
      leave(rest)
    end
  end

  SymbolRE     = /[\w\-\+\/\*\?\!]/
  SymbolREMany = /[\w\-\+\/\*\?\!]*/
  NumberRE     = /[\d\.]/
  NumberREMany = /[\d\.]*/

  class Main
    def initialize(text)
      smulti(:parse, /\s/)     {|c, rest| parse(rest)                }
      smulti(:parse, /\(/)     {|c, rest| @res = List.new(rest)      }
      smulti(:parse, /\"/)     {|c, rest| @res = String.new(rest)    }
      smulti(:parse, NumberRE) {|c, rest| @res = Number.new(rest, c) }
      smulti(:parse, SymbolRE) {|c, rest| @res = Symbol.new(rest, c) }
      parse(text)
    end

    def unwanted; return @res.unwanted(); end
    def value;    return @res.value()   ; end
  end
  
  class List < Base
    def initialize(*args)
      smulti(:parse, /\)/  ) {|s, rest| leave(rest) }
      smulti(:parse, /\s+/ ) {|s, rest| parse(rest) }
      smulti(:parse, //    ) {|s, rest|
        item = Main.new(rest)
        add(item.value)
        parse(item.unwanted)
      }
      super(*args)
    end

    def value; @list; end
  end
  
  class String < Base
    def initialize(*args)
      smulti(:parse, '\"')   {|s, rest| add_parse(s, rest) }
      smulti(:parse, '"' )   {|s, rest| leave(rest)        }
      smulti(:parse, /./ )   {|s, rest| add_parse(s, rest) }
      smulti(:parse, //  )   { throw "Couldn't parse string" }
      super(*args)
    end

    def value; @list.join(''); end
  end
  
  class Symbol < Base
    def initialize(*args)
      smulti(:parse, SymbolREMany) {|s, rest| add_leave(s, rest) }
      super(*args)
    end

    def value; @list.join('').to_sym; end
  end
  
  class Number < Base
    def initialize(*args)
      smulti(:parse, NumberREMany) {|s, rest| add_leave(s, rest) }
      super(*args)
    end

    def value; @list.join('').to_f; end
  end
end

class Object; def to_sexp; inspect(); end; end
class Symbol; def to_sexp; id2name(); end; end
class Array;  def to_sexp; "(#{map{|x| x.to_sexp }.join(' ')})"; end; end
class String; def parse_sexp; SExpressionParser::Main.new(self).value ; end; end
